@@ -2,6 +2,7 @@
 * rawhide - find files using pretty C expressions
 * https://raf.org/rawhide
 * https://github.com/raforg/rawhide
+* https://codeberg.org/raforg/rawhide
 *
 # Copyright (C) 1990 Ken Stauffer, 2022 raf <raf@raf.org>
 *
@@ -18,7 +19,7 @@
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <https://www.gnu.org/licenses/>.
 *
-* 20220330 raf <raf@raf.org>
+* 20221011 raf <raf@raf.org>
 */
 
 #define _GNU_SOURCE /* For FNM_EXTMATCH and FNM_CASEFOLD in <fnmatch.h> */
@@ -64,7 +65,7 @@ static void help_message(void)
 
 	printf("usage: %s [options] [path...]\n", RAWHIDE_PROG_NAME);
 	printf("options:\n");
-	printf("  -h --help    - Show this message, then exit\n");
+	printf("  -h --help    - Show this help message, then exit\n");
 	printf("  -V --version - Show the version message, then exit\n");
 	printf("  -N           - Don't read system-wide config (" RAWHIDE_CONF ")\n");
 	printf("  -n           - Don't read user-specific config (~" RAWHIDE_RC ")\n");
@@ -78,7 +79,7 @@ static void help_message(void)
 	printf("  -D           - Depth-first searching (contents before directory)\n");
 	printf("  -1           - Single filesystem (don't cross filesystem boundaries)\n");
 	printf("  -y           - Follow symlinks on the cmdline and in reference files\n");
-	printf("  -Y           - Follow symlinks while searching as well\n");
+	printf("  -Y           - Follow symlinks encountered while searching as well\n");
 	printf("\n");
 	printf("alternative action options:\n");
 	printf("  -x 'cmd %%s'  - Execute a shell command for each match (racy)\n");
@@ -189,6 +190,7 @@ static void help_message(void)
 	printf("Authors: Ken Stauffer, raf <raf@raf.org>\n");
 	printf("URL: https://raf.org/rawhide\n");
 	printf("GIT: https://github.org/raforg/rawhide\n");
+	printf("GIT: https://codeberg.org/raforg/rawhide\n");
 	printf("\n");
 	printf("Copyright (C) 1990 Ken Stauffer, 2022 raf <raf@raf.org>\n");
 	printf("\n");
@@ -200,6 +202,7 @@ static void help_message(void)
 	printf("for a particular purpose.\n");
 	printf("\n");
 	printf("Report bugs to raf <raf@raf.org>\n");
+
 	exit(EXIT_SUCCESS);
 }
 
@@ -214,6 +217,7 @@ Output the version message, then exit.
 static void version_message(void)
 {
 	printf("%s-%s\n", RAWHIDE_PROG_NAME, RAWHIDE_VERSION);
+
 	exit(EXIT_SUCCESS);
 }
 
@@ -222,7 +226,7 @@ static void version_message(void)
 
 static void debugf(const char *format, ...);
 
-Output a debug message to stderr.
+Output a cmdline debug message to stderr, if requested.
 
 */
 
@@ -243,22 +247,23 @@ static void debugf(const char *format, ...)
 
 /*
 
-static void load_program_file(void);
+static void load_program_file(char *fname);
 
-Compile the file expfname.
+Compile the given file. I/O errors are non-fatal.
 
 */
 
-static void load_program_file(void)
+static void load_program_file(char *fname)
 {
 	struct stat statbuf[1];
 
-	if (stat(expfname, statbuf) == -1 || (statbuf->st_mode & S_IFMT) != S_IFREG)
-		error("%s is not a file", ok(expfname));
-	else if (!(expfile = fopen(expfname, "r")))
-		errorsys("%s", ok(expfname));
+	if (stat(fname, statbuf) == -1 || !isreg(statbuf))
+		error("%s is not a file", ok(fname));
+	else if (!(expfile = fopen(fname, "r")))
+		errorsys("%s", ok(fname));
 	else
 	{
+		expfname = fname;
 		parse_program();
 		fclose(expfile);
 		expfile = NULL;
@@ -269,7 +274,7 @@ static void load_program_file(void)
 
 static void load_program_str(char *str);
 
-Compile str.
+Compile the given string.
 
 */
 
@@ -297,17 +302,82 @@ static int env_flag(char *envname)
 
 /*
 
+static void load_program_dir(const char *initdir, char *initpattern, llong pathbufsize);
+
+Compile the files (in lexicographic order) that are in the
+directory named in initdir. The initpattern buffer will be
+used to construct the glob pattern. The pathbufsize is the
+size of the two buffers.
+
+*/
+
+static void load_program_dir(const char *initdir, char *initpattern, llong pathbufsize)
+{
+	strlcpy(initpattern, initdir, pathbufsize);
+
+	if (strlcat(initpattern, "/*", pathbufsize) >= pathbufsize)
+		error("path is too long: %s/*", ok(initdir));
+	else
+	{
+		glob_t glob_state[1];
+		int i;
+
+		if (glob(initpattern, 0, NULL, glob_state) == 0)
+			for (i = 0; i < glob_state->gl_pathc; i++)
+				load_program_file(glob_state->gl_pathv[i]);
+
+		globfree(glob_state);
+	}
+}
+
+/*
+
+static void load_config_dir(const char *initfile, char *initdir, char *initpattern, llong pathbufsize);
+
+Compile the files (in lexicographic order) that are in the
+directory corresponding to the file named in initfile. The
+initdir and initpattern buffers will be used to construct
+the directory name (by appending ".d") and the glob pattern.
+The pathbufsize is the size of the buffers.
+
+*/
+
+static int load_config_dir(const char *initfile, char *initdir, char *initpattern, llong pathbufsize)
+{
+	struct stat statbuf[1];
+
+	strlcpy(initdir, initfile, pathbufsize);
+
+	if (strlcat(initdir, DOTDDIR, pathbufsize) >= pathbufsize)
+		error("path is too long: %s%s", ok(initfile), ok2(DOTDDIR));
+	else if (stat(initdir, statbuf) != -1)
+	{
+		if (!isdir(statbuf))
+			error("%s is not a directory", ok(initdir));
+		else
+		{
+			load_program_dir(initdir, initpattern, pathbufsize);
+
+			return 1;
+		}
+	}
+
+	return 0;
+}
+
+/*
+
 int main(int argc, char *argv[]);
 
 Parse the command line:
 
 The --help and --version options are supported (if first).
-The -f, -e, -x, -X, and -L options can only appear once.
+The -e, -x, -X, and -L options can only appear once.
 The -l, -0, -L, -x, -X, and -U options are mutually exclusive.
 
 Read /etc/rawhide.conf and /etc/rawhide.conf.d (unless -N).
 Read ~/.rhrc and ~/.rhrc.d (unless -n).
-Read -f file (if present).
+Read -f file/dir/stdin (if present).
 Read -e expr (if present).
 
 If no explicit -e expression is supplied, look among any
@@ -328,26 +398,25 @@ By default, matching entry names are output.
 The -l option and other options include more details.
 There are many options to affect the formatting.
 The -0 option outputs nul characters rather than newlines.
-The -L option outputs info in a user-defined format.
+The -L option outputs information in a user-defined format.
 The -x option executes a shell command for each
 matching file instead.
-The -X option is the same but executes each command
-from the matching entry's parent directory to minimize
-race conditions.
-The -U option unlinks matching entries instead, and must
-be supplied three times.
+The -X option is the same, but executes each command
+from the matching entry's parent directory so as to
+minimize race conditions.
+The -U option unlinks/removes/deletes matching entries instead,
+and must be supplied three times in order to take effect.
 
 */
 
 int main(int argc, char *argv[])
 {
-	char *opt_e, *opt_f, *initfile, *initdir, *initpattern, *endptr;
-	int opt_h, opt_V, opt_l, opt_r, opt_N, opt_n, opt_U;
+	char *opt_e, *initfile, *initdir, *initpattern, *endptr, **opt_f_list;
+	int opt_f, opt_h, opt_V, opt_l, opt_r, opt_N, opt_n, opt_U;
 	llong opt_m, opt_M, optarg_int, opt_e_expr = 0;
 	llong max_pathlen, pathbufsize;
-	glob_t glob_state[1];
-	int o, i, j, expr_index = 0, any = 0;
 	struct stat statbuf[1];
+	int o, i, expr_index = 0, any = 0, stdin_read = 0;
 
 	setlocale(LC_ALL, "");
 	prog_name = argv[0];
@@ -356,9 +425,10 @@ int main(int argc, char *argv[])
 
 	opt_h = 0;                  /* -h (help) */
 	opt_V = 0;                  /* -V (version) */
-	opt_N = 0;                  /* -N (suppress /etc/rawhide.conf)*/
-	opt_n = 0;                  /* -n (suppress ~/.rhrc) */
-	opt_f = NULL;               /* -f fname (read code from file/stdin) */
+	opt_N = 0;                  /* -N (suppress /etc/rawhide.conf and .d) */
+	opt_n = 0;                  /* -n (suppress ~/.rhrc and .d) */
+	opt_f = 0;                  /* Number of -f option arguments */
+	opt_f_list = NULL;          /* List of -f option arguments (file/dir/stdin) */
 	opt_e = NULL;               /* -e expr (read code from arg) */
 
 	opt_r = 0;                  /* -r (only search 1 level down) */
@@ -366,12 +436,12 @@ int main(int argc, char *argv[])
 	opt_M = -1;                 /* -M (max depth) */
 	attr.depth_first = 0;       /* -D (depth-first) */
 	attr.single_filesystem = 0; /* -1 (single filesystem) */
-	attr.follow_symlinks = 0;   /* -y -Y (follow symlinks) */
+	attr.follow_symlinks = 0;   /* -y -Y (follow symlinks: 1=supplied 2=encountered) */
 
 	attr.command = NULL;        /* -x/-X cmd (execute cmd) */
 	attr.local = 0;             /* cmd executed locally (-X) */
-	opt_U = 0;                  /* -U (delete - but tell me three times) */
-	attr.unlink = 0;            /* -UUU (unlink) */
+	opt_U = 0;                  /* -U (unlink - but tell me three times) */
+	attr.unlink = 0;            /* -UUU (unlink - I've been told three times) */
 
 	opt_l = 0;                  /* -l */
 	attr.dev_column = 0;        /* -d */
@@ -418,6 +488,17 @@ int main(int argc, char *argv[])
 	attr.fea_solaris_no_statinfo = env_flag("RAWHIDE_SOLARIS_EA_NO_STATINFO");
 	attr.fea_size = env_int("RAWHIDE_EA_SIZE", 1, -1, 0);
 
+	attr.test_cmd_max = env_int("RAWHIDE_TEST_CMD_MAX", 1, -1, -1);
+	attr.test_attr_format = env_flag("RAWHIDE_TEST_ATTR_FORMAT");
+	attr.test_chdir_failure = env_flag("RAWHIDE_TEST_CHDIR_FAILURE");
+	attr.test_fchdir_failure = env_flag("RAWHIDE_TEST_FCHDIR_FAILURE");
+	attr.test_fdopendir_failure = env_flag("RAWHIDE_TEST_FDOPENDIR_FAILURE");
+	attr.test_fstatat_failure = getenv("RAWHIDE_TEST_FSTATAT_FAILURE");
+	attr.test_invalid_date = env_flag("RAWHIDE_TEST_INVALID_DATE");
+	attr.test_openat_failure = env_flag("RAWHIDE_TEST_OPENAT_FAILURE");
+	attr.test_readlinkat_failure = env_flag("RAWHIDE_TEST_READLINKAT_FAILURE");
+	attr.test_readlinkat_too_long_failure = env_flag("RAWHIDE_TEST_READLINKAT_TOO_LONG_FAILURE");
+
 	/* Parse cmdline options */
 
 	rawhide_init();
@@ -465,10 +546,10 @@ int main(int argc, char *argv[])
 				if (!optarg || !*optarg)
 					fatal("missing -f option argument (see %s -h for help)", prog_name);
 
-				if (opt_f)
-					fatal("too many -f options");
+				if (!(opt_f_list = realloc(opt_f_list, ++opt_f * sizeof(*opt_f_list))))
+					fatalsys("out of memory");
 
-				opt_f = optarg;
+				opt_f_list[opt_f - 1] = optarg;
 
 				break;
 			}
@@ -767,7 +848,7 @@ int main(int argc, char *argv[])
 			{
 				/* Invalid options are reported with '?' and optopt set */
 
-				if (optopt && optopt != '?') /* macOS puts ? in optopt even when it's in optstring */
+				if (optopt && optopt != '?') /* macOS puts ? in optopt even though it's in optstring */
 					fatal("invalid option: -%c (see %s -h for help)", optopt, prog_name);
 
 				/* Otherwise, the ? option was supplied */
@@ -881,7 +962,7 @@ int main(int argc, char *argv[])
 	if (opt_V)
 		version_message();
 
-	/* For -X and "cmd".sh, remove cwd from $PATH (after debug config) */
+	/* For -X and "cmd".sh, remove cwd from $PATH (after debug config), affects -x as well */
 
 	if (remove_danger_from_path() == -1)
 		fatalsys("failed to remove cwd from path");
@@ -909,47 +990,11 @@ int main(int argc, char *argv[])
 		char *conf, *env;
 
 		conf = (geteuid() && (env = getenv("RAWHIDE_CONFIG")) && *env) ? env : RAWHIDE_CONF;
-		expfname = initfile;
 
-		if (strlcpy(expfname, conf, pathbufsize) >= pathbufsize)
-			error("path is too long: %s", ok(conf));
-		else
-		{
-			if (stat(expfname, statbuf) != -1)
-				load_program_file();
+		if (stat(conf, statbuf) != -1)
+			load_program_file(conf);
 
-			strlcpy(initdir, initfile, pathbufsize);
-
-			if (strlcat(initdir, DOTDDIR, pathbufsize) >= pathbufsize)
-				error("path is too long: %s%s", ok(initfile), ok2(DOTDDIR));
-			else if (stat(initdir, statbuf) != -1)
-			{
-				if ((statbuf->st_mode & S_IFMT) != S_IFDIR)
-					error("%s is not a directory", ok(initdir));
-				else
-				{
-					strlcpy(initpattern, initdir, pathbufsize);
-
-					if (strlcat(initpattern, "/*", pathbufsize) >= pathbufsize)
-						error("path is too long: %s/*", ok(initdir));
-					else
-					{
-						if (glob(initpattern, 0, NULL, glob_state) == 0)
-						{
-							for (j = 0; j < glob_state->gl_pathc; j++)
-							{
-								if (strlcpy(expfname, glob_state->gl_pathv[j], pathbufsize) >= pathbufsize)
-									error("path is too long: %s", ok(glob_state->gl_pathv[j]));
-								else
-									load_program_file();
-							}
-						}
-
-						globfree(glob_state);
-					}
-				}
-			}
-		}
+		load_config_dir(conf, initdir, initpattern, pathbufsize);
 	}
 
 	/* Load ~/.rhrc and ~/.rhrc.d (unless -n) */
@@ -964,52 +1009,71 @@ int main(int argc, char *argv[])
 		else if ((pwd = getpwuid(getuid())))
 			home = pwd->pw_dir;
 
-		expfname = initfile;
-		*expfname = '\0';
+		*initfile = '\0';
 
-		if (rc && strlcpy(expfname, rc, pathbufsize) >= pathbufsize)
+		if (rc && strlcpy(initfile, rc, pathbufsize) >= pathbufsize)
 			error("path is too long: %s", ok(rc));
-		else if (home && (strlcpy(expfname, home, pathbufsize) >= pathbufsize || strlcat(expfname, RAWHIDE_RC, pathbufsize) >= pathbufsize))
+		else if (home && (strlcpy(initfile, home, pathbufsize) >= pathbufsize || strlcat(initfile, RAWHIDE_RC, pathbufsize) >= pathbufsize))
 			error("path is too long: %s%s", ok(home), ok2(RAWHIDE_RC));
 		else
 		{
-			if (stat(expfname, statbuf) != -1)
-				load_program_file();
+			if (stat(initfile, statbuf) != -1)
+				load_program_file(initfile);
 
-			strlcpy(initdir, initfile, pathbufsize);
-
-			if (strlcat(initdir, DOTDDIR, pathbufsize) >= pathbufsize)
-				error("path is too long: %s%s", ok(initfile), ok2(DOTDDIR));
-			else if (stat(initdir, statbuf) != -1)
-			{
-				if ((statbuf->st_mode & S_IFMT) != S_IFDIR)
-					error("%s is not a directory", ok(initdir));
-				else
-				{
-					strlcpy(initpattern, initdir, pathbufsize);
-
-					if (strlcat(initpattern, "/*", pathbufsize) >= pathbufsize)
-						error("path is too long: %s/*", ok(initdir));
-					else
-					{
-						if (glob(initpattern, 0, NULL, glob_state) == 0)
-						{
-							for (j = 0; j < glob_state->gl_pathc; j++)
-							{
-								if (strlcpy(expfname, glob_state->gl_pathv[j], pathbufsize) >= pathbufsize)
-									error("path is too long: %s", ok(glob_state->gl_pathv[j]));
-								else
-									load_program_file();
-							}
-						}
-
-						globfree(glob_state);
-					}
-				}
-			}
+			load_config_dir(initfile, initdir, initpattern, pathbufsize);
 		}
 	}
 
+	/* Load the -f file-and-or-dir ("-" is stdin) */
+
+	for (i = 0; i < opt_f; ++i)
+	{
+		debug(("-f option: %s", opt_f_list[i]));
+
+		if (!strcmp(opt_f_list[i], "-"))
+		{
+			if (stdin_read)
+				fatal("invalid -f option argument: - (stdin can only be read once)");
+
+			stdin_read = 1;
+
+			expfname = "stdin";
+			expfile = stdin;
+			parse_program();
+			expfile = NULL;
+		}
+		else
+		{
+			if (stat(opt_f_list[i], statbuf) == -1)
+				fatalsys("invalid -f option argument: %s", ok(opt_f_list[i]));
+
+			if (isreg(statbuf))
+			{
+				if (!(expfile = fopen(opt_f_list[i], "r")))
+					fatalsys("%s", ok(opt_f_list[i]));
+
+				expfname = opt_f_list[i];
+				parse_program();
+				fclose(expfile);
+				expfile = NULL;
+
+				load_config_dir(opt_f_list[i], initdir, initpattern, pathbufsize);
+			}
+			else if (isdir(statbuf))
+				load_program_dir(opt_f_list[i], initpattern, pathbufsize);
+			else
+				fatal("invalid -f option argument: %s (not a file or directory)", ok(opt_f_list[i]));
+		}
+	}
+
+	/* Clean up */
+
+	if (opt_f_list)
+	{
+		free(opt_f_list);
+		opt_f_list = NULL;
+	}
+
 	free(initfile);
 	initfile = NULL;
 
@@ -1019,32 +1083,6 @@ int main(int argc, char *argv[])
 	free(initpattern);
 	initpattern = NULL;
 
-	/* Load the -f file ("-" is stdin) */
-
-	if (opt_f)
-	{
-		debug(("explicit -f option: %s", opt_f));
-
-		if (!strcmp(opt_f, "-"))
-		{
-			expfname = "stdin";
-			expfile = stdin;
-			parse_program();
-			expfile = NULL;
-		}
-		else
-		{
-			expfname = opt_f;
-
-			if (!(expfile = fopen(expfname, "r")))
-				fatalsys("%s", ok(expfname));
-
-			parse_program();
-			fclose(expfile);
-			expfile = NULL;
-		}
-	}
-
 	/* Parse the -e argument */
 
 	if (opt_e)
@@ -1056,7 +1094,7 @@ int main(int argc, char *argv[])
 		opt_e_expr = (startPC != opt_e_expr);
 	}
 
-	/* If no explicit expr, search arguments */
+	/* If no explicit -e expr, search remaining arguments for an implicit expr */
 
 	if (!opt_e_expr)
 	{
@@ -1188,7 +1226,7 @@ int main(int argc, char *argv[])
 	if (!(attr.search_stack = (point_t *)calloc(attr.max_depth + 1, sizeof(point_t))))
 		fatal("out of memory");
 
-	/* Find matches in the given directories or the current directory */
+	/* Find matches in the given directories (or the current working directory) */
 
 	for (; optind < argc; optind++)
 	{
@@ -1208,6 +1246,7 @@ int main(int argc, char *argv[])
 	}
 
 	free(attr.search_stack);
+
 	exit(attr.exit_status);
 }
 
