@@ -2,6 +2,7 @@
 * rawhide - find files using pretty C expressions
 * https://raf.org/rawhide
 * https://github.com/raforg/rawhide
+* https://codeberg.org/raforg/rawhide
 *
 * Copyright (C) 1990 Ken Stauffer, 2022 raf <raf@raf.org>
 *
@@ -18,7 +19,7 @@
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <https://www.gnu.org/licenses/>.
 *
-* 20220330 raf <raf@raf.org>
+* 20221011 raf <raf@raf.org>
 */
 
 #define _GNU_SOURCE /* For FNM_EXTMATCH and FNM_CASEFOLD in <fnmatch.h> */
@@ -69,14 +70,16 @@
 #include <sys/extattr.h>
 #endif
 
+#ifdef HAVE_ATTR
+#include <e2p/e2p.h>
+#include <ext2fs/ext2_fs.h>
+#endif
+
 #include "rh.h"
 #include "rhdir.h"
 #include "rherr.h"
 #include "rhstr.h"
 
-#define isdir(statbuf) (((statbuf)->st_mode & S_IFMT) == S_IFDIR)
-#define islink(statbuf) (((statbuf)->st_mode & S_IFMT) == S_IFLNK)
-
 /* Operators */
 
 void c_le(llong i)     { Stack[SP - 2] = Stack[SP - 2] <= Stack[SP - 1]; SP--; }
@@ -222,6 +225,51 @@ void c_prune(llong i)   { Stack[SP++] = 1; attr.prune = attr.pruned = 1; }
 void c_trim(llong i)    { Stack[SP++] = 1; attr.prune = 1; }
 void c_exit(llong i)    { Stack[SP++] = 1; attr.exit = 1; }
 
+unsigned long get_attr(void)
+{
+	#ifdef HAVE_ATTR
+	if (!attr.attr_done)
+	{
+		attr.attr_done = 1;
+		fgetflags(attr.fpath, &attr.attr);
+	}
+	#endif
+
+	return attr.attr;
+}
+
+unsigned long get_proj(void)
+{
+	#ifdef HAVE_ATTR
+	if (!attr.proj_done)
+	{
+		attr.proj_done = 1;
+		fgetproject(attr.fpath, &attr.proj);
+	}
+	#endif
+
+	return attr.proj;
+}
+
+unsigned long get_gen(void)
+{
+	#ifdef HAVE_ATTR
+	if (!attr.gen_done)
+	{
+		attr.gen_done = 1;
+		fgetversion(attr.fpath, &attr.gen);
+	}
+	#endif
+
+	return attr.gen;
+}
+
+#ifdef HAVE_ATTR
+void c_attr(llong i) { Stack[SP++] = get_attr(); }
+void c_proj(llong i) { Stack[SP++] = get_proj(); }
+void c_gen(llong i)  { Stack[SP++] = get_gen(); }
+#endif
+
 void c_strlen(llong i)
 {
 	int len;
@@ -277,21 +325,24 @@ static char *c_basename(void)
 	return tail;
 }
 
-/* Read the symlink target path */
+/* Read the current candidate symlink target path */
 
 char *read_symlink(void)
 {
-	ssize_t nbytes;
+	ssize_t nbytes = 0;
 
 	if (!attr.ftarget_done)
 	{
 		if (!attr.ftarget && !(attr.ftarget = malloc(attr.fpath_size)))
 			fatalsys("out of memory");
 
-		if ((nbytes = readlinkat(attr.parent_fd, (attr.parent_fd == AT_FDCWD) ? attr.fpath : attr.basename, attr.ftarget, attr.fpath_size)) == -1)
+		if (attr.test_readlinkat_failure)
+			errno = EPERM;
+
+		if (attr.test_readlinkat_failure || (nbytes = readlinkat(attr.parent_fd, (attr.parent_fd == AT_FDCWD) ? attr.fpath : attr.basename, attr.ftarget, attr.fpath_size)) == -1)
 			fatalsys("readlinkat %s", ok(attr.fpath));
 
-		if (nbytes == attr.fpath_size)
+		if (attr.test_readlinkat_too_long_failure || nbytes == attr.fpath_size)
 			fatal("readlinkat %s: target is too long", ok(attr.fpath));
 
 		attr.ftarget[nbytes] = '\0';
@@ -308,7 +359,7 @@ char *read_symlink(void)
 
 Glob pattern matching.
 
-The i parameter is an index into Strbuf[].
+The i parameter is an index into Strbuf.
 The string contained there is the nul-terminated
 string that is the pattern (e.g., "*.bak" without
 the double quotes).
@@ -342,6 +393,8 @@ static pcre2_code *pcre2_compile_checked(const char *pattern, uint32_t options)
 		fatal("invalid regex %s at offset %d: %s", ok(pattern), (int)error_offset, ok2((char *)error_buffer));
 	}
 
+	pcre2_jit_compile(re, PCRE2_JIT_COMPLETE);
+
 	return re;
 }
 
@@ -421,7 +474,7 @@ static int rematch(const char *pattern, const char *subject, uint32_t options)
 	options |= PCRE2_EXTENDED;          /* Ignore whitespace and # comments (except in character classes) (like /x) */
 	options |= PCRE2_EXTENDED_MORE;     /* Ignore space and tab inside character classes as well (like /xx) */
 	options |= PCRE2_NO_AUTO_CAPTURE;   /* Prevent automatic numbered capturing parentheses (like /n) */
-	options |= PCRE2_NEVER_BACKSLASH_C; /* Prevents the use of the \C escape in patterns */
+	options |= PCRE2_NEVER_BACKSLASH_C; /* Prevent the use of the \C escape in patterns */
 
 	if (attr.utf)
 		options |= PCRE2_UTF | PCRE2_MATCH_INVALID_UTF; /* Assume UTF-8 patterns and subject text */
@@ -953,7 +1006,7 @@ char *get_ea(int want)
 
 		attr.fea_done = 1;
 
-		/* Symlinks can't have EAs on Solaris so handle symlink following here */
+		/* Symlinks can't have EAs on Solaris so handle symlink-following here */
 
 		if (islink(attr.statbuf) && !following_symlinks())
 			return NULL;
@@ -1133,7 +1186,7 @@ int has_real_acl(void)
 	if (strstr(acl, "user::"))
 		return strstr(acl, "mask::") || (get_ea(0) && attr.fea_ok && strstr(attr.fea, "system.posix_acl_access: "));
 
-	/* NFSv4 ACLs (approximation) */
+	/* NFSv4 ACLs */
 
 	if (strstr(acl, "owner@:"))
 		return strstr(acl, "user:") || strstr(acl, "group:");
@@ -1233,6 +1286,49 @@ void r_ctime(llong i)   { check_reference(i, "ctime");   Stack[SP++] = RefFile[i
 void r_type(llong i)    { check_reference(i, "type");    Stack[SP++] = RefFile[i].statbuf->st_mode & S_IFMT; }
 void r_perm(llong i)    { check_reference(i, "perm");    Stack[SP++] = RefFile[i].statbuf->st_mode & ~S_IFMT; }
 
+#ifdef HAVE_ATTR
+
+void r_attr(llong i)
+{
+	check_reference(i, "attr");
+
+	if (!RefFile[i].attr_done)
+	{
+		RefFile[i].attr_done = 1;
+		fgetflags(Strbuf + RefFile[i].fpathi, &RefFile[i].attr);
+	}
+
+	Stack[SP++] = RefFile[i].attr;
+}
+
+void r_proj(llong i)
+{
+	check_reference(i, "proj");
+
+	if (!RefFile[i].proj_done)
+	{
+		RefFile[i].proj_done = 1;
+		fgetproject(Strbuf + RefFile[i].fpathi, &RefFile[i].proj);
+	}
+
+	Stack[SP++] = RefFile[i].proj;
+}
+
+void r_gen(llong i)
+{
+	check_reference(i, "gen");
+
+	if (!RefFile[i].gen_done)
+	{
+		RefFile[i].gen_done = 1;
+		fgetversion(Strbuf + RefFile[i].fpathi, &RefFile[i].gen);
+	}
+
+	Stack[SP++] = RefFile[i].gen;
+}
+
+#endif
+
 /* Symlink targets */
 
 void prepare_target(void)
@@ -1332,6 +1428,11 @@ char *instruction_name(void (*func)(llong))
 		(func == c_atime) ? "atime" :
 		(func == c_mtime) ? "mtime" :
 		(func == c_ctime) ? "ctime" :
+		#ifdef HAVE_ATTR
+		(func == c_attr) ? "cattr" :
+		(func == c_proj) ? "cproj" :
+		(func == c_gen) ? "cgen" :
+		#endif
 		(func == c_depth) ? "depth" :
 		(func == c_prune) ? "prune" :
 		(func == c_trim) ? "trim" :
@@ -1397,6 +1498,11 @@ char *instruction_name(void (*func)(llong))
 		(func == r_atime) ? "ratime" :
 		(func == r_mtime) ? "rmtime" :
 		(func == r_ctime) ? "rctime" :
+		#ifdef HAVE_ATTR
+		(func == r_attr) ? "rattr" :
+		(func == r_proj) ? "rproj" :
+		(func == r_gen) ? "rgen" :
+		#endif
 		(func == r_strlen) ? "rstrlen" :
 		(func == r_type) ? "rtype" :
 		(func == r_perm) ? "rperm" :
