@@ -2,6 +2,7 @@
 * rawhide - find files using pretty C expressions
 * https://raf.org/rawhide
 * https://github.com/raforg/rawhide
+* https://codeberg.org/raforg/rawhide
 *
 * Copyright (C) 1990 Ken Stauffer, 2022 raf <raf@raf.org>
 *
@@ -18,7 +19,7 @@
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <https://www.gnu.org/licenses/>.
 *
-* 20220330 raf <raf@raf.org>
+* 20221011 raf <raf@raf.org>
 */
 
 #define _GNU_SOURCE /* For FNM_EXTMATCH and FNM_CASEFOLD in <fnmatch.h> */
@@ -37,8 +38,7 @@
 #include "rhcmds.h"
 #include "rherr.h"
 
-char *prog_name;                     /* Name of the program for messages */
-runtime_t attr;                      /* Configuration and runtime state */
+char *prog_name;                     /* Name of this program for messages */
 
 symbol_t *symbols;                   /* Symbol table */
 symbol_t *tokensym;                  /* Current token symbol */
@@ -49,19 +49,21 @@ instr_t Program[MAX_PROGRAM_SIZE];   /* Program instructions */
 llong PC;                            /* Program counter */
 llong startPC;                       /* Initial program counter */
 
-llong Stack[MAX_STACK_SIZE + 3];     /* Stack */
-llong SP;                            /* Stack pointer */
-llong FP;                            /* Frame pointer */
-
 char Strbuf[MAX_DATA_SIZE];          /* Storage for string literals */
 llong strfree = 0;                   /* Index into Strbuf for the next string literal */
 
 reffile_t RefFile[MAX_REFFILE_SIZE]; /* Storage for reference files */
-llong reffree = 0;                   /* Index into RefFile for the next one */
+llong reffree = 0;                   /* Index into RefFile for the next reference file */
 
-char *expstr;   /* The -e option argument */
-char *expfname; /* The -f option filename or current config file */
-FILE *expfile;  /* FILE handle for expfname */
+llong Stack[MAX_STACK_SIZE + 3];     /* Stack */
+llong SP;                            /* Stack pointer */
+llong FP;                            /* Frame pointer */
+
+runtime_t attr;                      /* Configuration and runtime state */
+
+char *expstr;   /* The -e option argument (search criteria expression) */
+char *expfname; /* The -f option argument (filename) or current config file */
+FILE *expfile;  /* FILE handle corresponding to the -f option argument or current config file */
 
 #ifndef S_IFDOOR
 #define S_IFDOOR 0150000
@@ -91,6 +93,14 @@ static symbol_t init_syms[] =
 	{ "mtime",   FIELD, 0, c_mtime,   NULL },
 	{ "ctime",   FIELD, 0, c_ctime,   NULL },
 
+	/* Linux ext2-style file attributes */
+
+#ifdef HAVE_ATTR
+	{ "attr",    FIELD, 0, c_attr,   NULL },
+	{ "proj",    FIELD, 0, c_proj,   NULL },
+	{ "gen",     FIELD, 0, c_gen,    NULL },
+#endif
+
 	/* Miscellaneous functions and control flow */
 
 	{ "nouser",     FIELD, 0, c_nouser,     NULL },
@@ -187,6 +197,11 @@ static symbol_t init_syms[] =
 	{ ".atime",    REFFILE, 0, r_atime,   NULL },
 	{ ".mtime",    REFFILE, 0, r_mtime,   NULL },
 	{ ".ctime",    REFFILE, 0, r_ctime,   NULL },
+#ifdef HAVE_ATTR
+	{ ".attr",     REFFILE, 0, r_attr,    NULL },
+	{ ".proj",     REFFILE, 0, r_proj,    NULL },
+	{ ".gen",      REFFILE, 0, r_gen,     NULL },
+#endif
 	{ ".strlen",   REFFILE, 0, r_strlen,  NULL },
 
 	{ ".inode",    REFFILE, 0, r_ino,     NULL },
@@ -197,6 +212,11 @@ static symbol_t init_syms[] =
 	{ ".accessed", REFFILE, 0, r_atime,   NULL },
 	{ ".modified", REFFILE, 0, r_mtime,   NULL },
 	{ ".changed",  REFFILE, 0, r_ctime,   NULL },
+#ifdef HAVE_ATTR
+	{ ".attribute",  REFFILE, 0, r_attr,  NULL },
+	{ ".project",    REFFILE, 0, r_proj,  NULL },
+	{ ".generation", REFFILE, 0, r_gen,   NULL },
+#endif
 	{ ".len",      REFFILE, 0, r_strlen,  NULL },
 
 	/* Pattern modifiers */
@@ -368,6 +388,37 @@ symbol_t *locate_symbol(char *name)
 
 /*
 
+symbol_t *locate_patmod_prefix(char *name);
+
+Search for a pattern modifier symbol in the symbol table.
+The name parameter is hopefully a unique prefix of a symbol,
+but it is not a complete symbol.
+
+*/
+
+symbol_t *locate_patmod_prefix(char *name)
+{
+	symbol_t *s, *match = NULL;
+
+	for (s = symbols; s; s = s->next)
+	{
+		if (s->type != PATMOD)
+			continue;
+
+		if (!strncmp(name, s->name, strlen(name)))
+		{
+			if (match)
+				return NULL; /* Not a unique prefix */
+
+			match = s;
+		}
+	}
+
+	return match;
+}
+
+/*
+
 int rawhide_instruction(void (*func)(llong), llong value);
 
 Append a rawhide instruction to the Program array.
@@ -393,7 +444,7 @@ llong rawhide_execute(void);
 
 Execute the program stored in Program.
 Each element of Program contains a pointer to a function.
-The program is NULL terminated.
+The program is NULL-terminated.
 Returns the value of the expression.
 
 */
