@@ -2,6 +2,7 @@
 * rawhide - find files using pretty C expressions
 * https://raf.org/rawhide
 * https://github.com/raforg/rawhide
+* https://codeberg.org/raforg/rawhide
 *
 * Copyright (C) 1990 Ken Stauffer, 2022 raf <raf@raf.org>
 *
@@ -18,7 +19,7 @@
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <https://www.gnu.org/licenses/>.
 *
-* 20220330 raf <raf@raf.org>
+* 20221011 raf <raf@raf.org>
 */
 
 #ifndef RAWHIDE_RHDATA_H
@@ -28,6 +29,7 @@ void rawhide_init(void);
 void rawhide_finish(void);
 symbol_t *insert_symbol(char *name, int toktype, llong val);
 symbol_t *locate_symbol(char *name);
+symbol_t *locate_patmod_prefix(char *name);
 int rawhide_instruction(void (*func)(llong), llong value);
 llong rawhide_execute(void);
 
