@@ -2,6 +2,7 @@
 * rawhide - find files using pretty C expressions
 * https://raf.org/rawhide
 * https://github.com/raforg/rawhide
+* https://codeberg.org/raforg/rawhide
 *
 * Copyright (C) 1990 Ken Stauffer, 2022 raf <raf@raf.org>
 *
@@ -18,10 +19,10 @@
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <https://www.gnu.org/licenses/>.
 *
-* 20220330 raf <raf@raf.org>
+* 20221011 raf <raf@raf.org>
 */
 
-#define _GNU_SOURCE /* For FNM_EXTMATCH and FNM_CASEFOLD in <fnmatch.h> */
+#define _GNU_SOURCE /* For FNM_EXTMATCH and FNM_CASEFOLD in <fnmatch.h> and wcswidth() in <wchar.h> */
 #define _FILE_OFFSET_BITS 64 /* For 64-bit off_t on 32-bit systems (Not AIX) */
 
 #include <stdlib.h>
@@ -36,6 +37,7 @@
 #include <errno.h>
 #include <ctype.h>
 #include <fcntl.h>
+#include <wchar.h>
 #include <pwd.h>
 #include <grp.h>
 #include <sys/stat.h>
@@ -54,25 +56,16 @@
 #include <sys/mkdev.h>
 #endif
 
+#ifdef HAVE_ATTR
+#include <ext2fs/ext2_fs.h>
+#endif
+
 #include "rh.h"
 #include "rhdata.h"
 #include "rhcmds.h"
 #include "rhdir.h"
 #include "rhstr.h"
 #include "rherr.h"
-#include "rhjson.h"
-
-#define isdir(statbuf)  (((statbuf)->st_mode & S_IFMT) == S_IFDIR)
-#define isblk(statbuf)  (((statbuf)->st_mode & S_IFMT) == S_IFBLK)
-#define ischr(statbuf)  (((statbuf)->st_mode & S_IFMT) == S_IFCHR)
-#define islink(statbuf) (((statbuf)->st_mode & S_IFMT) == S_IFLNK)
-#define issock(statbuf) (((statbuf)->st_mode & S_IFMT) == S_IFSOCK)
-#define isfifo(statbuf) (((statbuf)->st_mode & S_IFMT) == S_IFIFO)
-#ifdef S_IFDOOR
-#define isdoor(statbuf) (((statbuf)->st_mode & S_IFMT) == S_IFDOOR)
-#else
-#define isdoor(statbuf) (0)
-#endif
 
 #ifdef NDEBUG
 #define debug(args)
@@ -91,7 +84,7 @@ static char cwdbuf[CMDBUFSIZE];
 
 static void debugf(const char *format, ...);
 
-Print a debug message to stderr.
+Output a traversal debug message to stderr, if requested.
 
 */
 
@@ -113,7 +106,7 @@ static void debugf(const char *format, ...)
 
 static void debug_extraf(const char *format, ...);
 
-Print an extra debug message to stderr.
+Output an extra traversal debug message to stderr, if requested.
 
 */
 
@@ -185,6 +178,12 @@ static void caches_init(void)
 	attr.fea_ok = 0;
 	attr.fea_selinux = NULL;
 	attr.fea_real = 0;
+	attr.attr_done = 0;
+	attr.attr = 0;
+	attr.proj_done = 0;
+	attr.proj = 0;
+	attr.gen_done = 0;
+	attr.gen = 0;
 }
 
 /*
@@ -245,13 +244,104 @@ int following_symlinks(void)
 
 /*
 
-static void rawhide_traverse(char *nul_posp, int parent_fd, char *basename);
+static int fcntl_set_fdflag(int fd, int flag);
+
+Shorthand for setting the file descriptor flag, flag, on the file
+descriptor fd, using fcntl. All other file descriptor flags are
+unaffected. On success, returns 0. On error, returns -1 with errno
+set by fcntl with F_GETFD or F_SETFD as the command. The only
+file descriptor flag at time of writing is FD_CLOEXEC.
+
+*/
+
+static int fcntl_set_fdflag(int fd, int flag)
+{
+	int flags;
+
+	if ((flags = fcntl(fd, F_GETFD, 0)) == -1)
+		return -1;
+
+	return fcntl(fd, F_SETFD, flags | flag);
+}
+
+/*
+
+static ssize_t wcoffset(const char *str);
+
+When a string contains multi-byte UTF-8-encoded characters,
+the length of the string is more than the display width.
+Return the difference between the two. If any errors are
+encountered (e.g. invalid characters for the locale,
+or out of memory), an offset of zero is returned.
+An internal dynamically allocated buffer might be needed.
+To deallocate it when finished, call with a NULL argument.
+
+*/
+
+static ssize_t wcoffset(const char *str)
+{
+	static wchar_t *wcs = NULL;
+	static size_t wcssize = 0;
+
+	ssize_t num_wchars;
+	size_t num_bytes;
+	int width;
+
+	/* Release the long-lived wide char buffer when finished */
+
+	if (!str)
+	{
+		if (wcs)
+		{
+			free(wcs);
+			wcs = NULL;
+			wcssize = 0;
+		}
+
+		return 0;
+	}
+
+	/* If there are no multi-byte characters, return a zero offset */
+
+	num_wchars = (ssize_t)mbstowcs(NULL, str, 0);
+	num_bytes = strlen(str);
+
+	if (num_wchars == -1 || num_wchars == num_bytes)
+		return 0;
+
+	/* Allocate a long-lived wide char buffer (grows as needed) */
+
+	if (wcssize < num_wchars + 1)
+	{
+		wchar_t *new_wcs = realloc(wcs, sizeof(wchar_t) * (num_wchars + 1));
+
+		if (!new_wcs)
+			return 0;
+
+		wcs = new_wcs;
+		wcssize = num_wchars;
+	}
+
+	/* Convert str to wide chars, then determine its display width */
+
+	mbstowcs(wcs, str, num_wchars);
+
+	width = wcswidth(wcs, num_wchars);
+
+	/* The offset is the difference between the bytes and the width */
+
+	return (width == -1) ? 0 : num_bytes - width;
+}
+
+/*
+
+static void rawhide_traverse(size_t nul_posi, int parent_fd, char *basename);
 
 Traverse the directory tree, evaluating search criteria, and calling the
 visit function, for each match.
 
-The nul_posp parameter points to the nul byte at the end of attr.fpath.
-There can be a trailing "/" before nul_posp (but usually not).
+The nul_posi parameter is the offset to the nul byte at the end of attr.fpath.
+There can be a trailing "/" before nul_posi (but usually not).
 
 The parent_fd parameter is a file descriptor for the current entry's parent
 directory.
@@ -264,7 +354,7 @@ race conditions. When called by rawhide_search() they are AT_FDCWD and NULL.
 
 */
 
-static int rawhide_traverse(char *nul_posp, int parent_fd, char *basename)
+static int rawhide_traverse(size_t nul_posi, int parent_fd, char *basename)
 {
 	int dir_fd;
 	DIR *dir = NULL;
@@ -274,9 +364,8 @@ static int rawhide_traverse(char *nul_posp, int parent_fd, char *basename)
 	int save_followed = 0;
 	int rc = 0, i;
 	char *name;
-	char *env;
 
-	debug(("rawhide_traverse(fpath=%s, offset=%d, parent_fd=%d, basename=%s, depth=%d)", attr.fpath, (int)(nul_posp - attr.fpath), parent_fd, (basename) ? basename : "N/A", attr.depth));
+	debug(("rawhide_traverse(fpath=%s, offset=%d, parent_fd=%d, basename=%s, depth=%d)", attr.fpath, (int)nul_posi, parent_fd, (basename) ? basename : "N/A", attr.depth));
 
 	/* Check the depth (Can't happen: depth <= max_depth always) */
 
@@ -296,13 +385,10 @@ static int rawhide_traverse(char *nul_posp, int parent_fd, char *basename)
 
 	debug(("fstatat(parent_fd=%d, path=%s)", parent_fd, name));
 
-	if ((env = getenv("RAWHIDE_TEST_FSTATAT_FAILURE")))
-	{
-		debug(("RAWHIDE_TEST_FSTATAT_FAILURE=%s", env));
+	if (attr.test_fstatat_failure && !strcmp(attr.test_fstatat_failure, attr.fpath))
 		errno = EPERM;
-	}
 
-	if ((env && !strcmp(env, attr.fpath)) || fstatat(parent_fd, name, attr.statbuf, AT_SYMLINK_NOFOLLOW) == -1)
+	if ((attr.test_fstatat_failure && !strcmp(attr.test_fstatat_failure, attr.fpath)) || fstatat(parent_fd, name, attr.statbuf, AT_SYMLINK_NOFOLLOW) == -1)
 		return errorsys("fstatat %s", ok(attr.fpath));
 
 	debug(("fstatat: basename %s type %o perm %o uid %d gid %d size %d", (basename) ? basename : "N/A", attr.statbuf->st_mode & S_IFMT, attr.statbuf->st_mode & ~S_IFMT, attr.statbuf->st_uid, attr.statbuf->st_gid, attr.statbuf->st_size));
@@ -379,13 +465,10 @@ static int rawhide_traverse(char *nul_posp, int parent_fd, char *basename)
 
 		debug(("openat(parent_fd=%d, path=%s)", parent_fd, name));
 
-		if ((env = getenv("RAWHIDE_TEST_OPENAT_FAILURE")))
-		{
-			debug(("RAWHIDE_TEST_OPENAT_FAILURE=%s", env));
+		if (attr.test_openat_failure)
 			errno = EPERM;
-		}
 
-		if ((env && *env == '1') || (dir_fd = openat(parent_fd, name, O_RDONLY)) == -1)
+		if (attr.test_openat_failure || (dir_fd = openat(parent_fd, name, O_RDONLY)) == -1)
 			return errorsys("%s", ok(attr.fpath));
 
 		debug(("openat: dir_fd %d", dir_fd));
@@ -398,13 +481,10 @@ static int rawhide_traverse(char *nul_posp, int parent_fd, char *basename)
 
 		debug(("fdopendir(dir_fd=%d)", dir_fd));
 
-		if ((env = getenv("RAWHIDE_TEST_FDOPENDIR_FAILURE")))
-		{
-			debug(("RAWHIDE_TEST_FDOPENDIR_FAILURE=%s", env));
+		if (attr.test_fdopendir_failure)
 			errno = EPERM;
-		}
 
-		if ((env && *env == '1') || !(dir = fdopendir(dir_fd)))
+		if (attr.test_fdopendir_failure || !(dir = fdopendir(dir_fd)))
 		{
 			close(dir_fd);
 
@@ -413,15 +493,25 @@ static int rawhide_traverse(char *nul_posp, int parent_fd, char *basename)
 
 		/* Ensure that there's a trailing "/" */
 
-		post_slash_nul_posi = nul_posp - attr.fpath;
+		post_slash_nul_posi = nul_posi;
 
 		if (attr.fpath[post_slash_nul_posi - 1] != '/')
 		{
 			if (post_slash_nul_posi + 1 >= attr.fpath_size)
 			{
-				closedir(dir); /* This closes pid_fd */
+				/* This shouldn't happen (unless we've crossed a filesystem boundary) */
 
-				return error("path is too long: %s/", ok(attr.fpath));
+				char *new_fpath = realloc(attr.fpath, attr.fpath_size * 2);
+
+				if (!new_fpath)
+				{
+					closedir(dir); /* This closes pid_fd */
+
+					return error("path is too long (out of memory): %s/", ok(attr.fpath));
+				}
+
+				attr.fpath = new_fpath;
+				attr.fpath_size *= 2;
 			}
 
 			attr.fpath[post_slash_nul_posi++] = '/';
@@ -445,16 +535,31 @@ static int rawhide_traverse(char *nul_posp, int parent_fd, char *basename)
 
 			if ((len = strlcpy(attr.fpath + post_slash_nul_posi, entry->d_name, remaining)) >= remaining)
 			{
-				attr.fpath[post_slash_nul_posi] = '\0';
-				error("path is too long: %s%s", ok(attr.fpath), ok2(entry->d_name));
-				rc = -1;
+				/* This shouldn't happen (unless we've crossed a filesystem boundary) */
 
-				continue;
+				#define max(a, b) ((a) > (b) ? (a) : (b))
+				size_t new_fpath_size = max(attr.fpath_size * 2, post_slash_nul_posi + len + 1);
+				#undef max
+				char *new_fpath = realloc(attr.fpath, new_fpath_size);
+
+				if (!new_fpath)
+				{
+					attr.fpath[post_slash_nul_posi] = '\0';
+					error("path is too long (out of memory): %s%s", ok(attr.fpath), ok2(entry->d_name));
+					rc = -1;
+
+					continue;
+				}
+
+				attr.fpath = new_fpath;
+				attr.fpath_size = new_fpath_size;
+				remaining = attr.fpath_size - post_slash_nul_posi;
+				strlcpy(attr.fpath + post_slash_nul_posi, entry->d_name, remaining);
 			}
 
 			debug(("dir entry %s", attr.fpath));
 
-			if (rawhide_traverse(attr.fpath + post_slash_nul_posi + len, (dir_fd == -1) ? AT_FDCWD : dir_fd, attr.fpath + post_slash_nul_posi) == -1)
+			if (rawhide_traverse(post_slash_nul_posi + len, (dir_fd == -1) ? AT_FDCWD : dir_fd, attr.fpath + post_slash_nul_posi) == -1)
 				rc = -1;
 		}
 
@@ -466,7 +571,7 @@ static int rawhide_traverse(char *nul_posp, int parent_fd, char *basename)
 
 		/* Remove the trailing slash and directory entry name */
 
-		*nul_posp = '\0';
+		attr.fpath[nul_posi] = '\0';
 	}
 
 	/* Test this entry and respond (if searching depth-first) */
@@ -577,7 +682,7 @@ int rawhide_search(char *fpath)
 
 	/* Search */
 
-	rc = rawhide_traverse(attr.fpath + len, AT_FDCWD, NULL);
+	rc = rawhide_traverse(len, AT_FDCWD, NULL);
 
 	/* Cleanup */
 
@@ -611,6 +716,8 @@ int rawhide_search(char *fpath)
 		attr.ttybuf = NULL;
 	}
 
+	wcoffset(NULL);
+
 	return rc;
 }
 
@@ -619,7 +726,7 @@ int rawhide_search(char *fpath)
 void visitf_default(void);
 
 The default action for matching files (without -l).
-It prints the name.
+It outputs the name.
 
 */
 
@@ -671,10 +778,19 @@ void visitf_default(void)
 	{
 		ssize_t nbytes;
 
-		if ((nbytes = readlinkat(attr.parent_fd, (attr.parent_fd == AT_FDCWD) ? attr.fpath : attr.basename, linkbuf, CMDBUFSIZE)) == -1)
+		if (attr.test_readlinkat_failure)
+			errno = EPERM;
+
+		if (attr.test_readlinkat_failure || (nbytes = readlinkat(attr.parent_fd, (attr.parent_fd == AT_FDCWD) ? attr.fpath : attr.basename, linkbuf, CMDBUFSIZE)) == -1)
+		{
+			attr.exit_status = EXIT_FAILURE;
 			errorsys("readlinkat %s", ok(attr.fpath));
-		else if (nbytes == CMDBUFSIZE)
+		}
+		else if (attr.test_readlinkat_too_long_failure || nbytes == CMDBUFSIZE)
+		{
+			attr.exit_status = EXIT_FAILURE;
 			error("readlinkat %s: target is too long", ok(attr.fpath));
+		}
 		else
 		{
 			linkbuf[nbytes] = '\0';
@@ -913,14 +1029,14 @@ static char *ftypecode(struct stat *statbuf)
 
 /*
 
-const char *ytypecode(struct stat *statbuf);
+static const char *ytypecode(struct stat *statbuf);
 
 Return a pointer to a static buffer containing the file type code in ls -l
 format, but with f for regular files (for -L %y and %Y).
 
 */
 
-const char *ytypecode(struct stat *statbuf)
+static const char *ytypecode(struct stat *statbuf)
 {
 	char *code = ftypecode(statbuf);
 
@@ -969,13 +1085,13 @@ const char *modestr(struct stat *statbuf)
 
 /*
 
-const char *aclea(void);
+static const char *aclea(void);
 
 Return a pointer to a static buffer containing the ACL/EA indicator.
 
 */
 
-const char *aclea(void)
+static const char *aclea(void)
 {
 	int acl = has_real_acl();
 	int ea = has_real_ea();
@@ -987,7 +1103,7 @@ const char *aclea(void)
 
 void visitf_long(void);
 
-The -l action for matching files. It prints the name and other details. The
+The -l action for matching files. It outputs the name and other details. The
 potential columns included are:
 
   dev, ino, blksize, blocks, space, mode (type+perm+aclea), nlink,
@@ -1171,8 +1287,10 @@ void visitf_long(void)
 		}
 		else
 		{
-			if ((w = ssnprintf(buf + pos, CMDBUFSIZE - pos, "%-*s", attr.user_column_width, pwd->pw_name)) > attr.user_column_width)
-				attr.user_column_width = w;
+			ssize_t off = wcoffset(pwd->pw_name);
+
+			if ((w = ssnprintf(buf + pos, CMDBUFSIZE - pos, "%-*s", attr.user_column_width + off, pwd->pw_name)) - off > attr.user_column_width)
+				attr.user_column_width = w - off;
 		}
 
 		pos += w;
@@ -1194,8 +1312,10 @@ void visitf_long(void)
 		}
 		else
 		{
-			if ((w = ssnprintf(buf + pos, CMDBUFSIZE - pos, "%-*s", attr.group_column_width, grp->gr_name)) > attr.group_column_width)
-				attr.group_column_width = w;
+			ssize_t off = wcoffset(grp->gr_name);
+
+			if ((w = ssnprintf(buf + pos, CMDBUFSIZE - pos, "%-*s", attr.group_column_width + off, grp->gr_name)) - off > attr.group_column_width)
+				attr.group_column_width = w - off;
 		}
 
 		pos += w;
@@ -1311,7 +1431,9 @@ int interpolate_command(const char *srccmd, char *command, int cmdbufsize)
 
 	src = srccmd;
 	dst = command;
-	cmdbufsize = env_int("RAWHIDE_TEST_CMD_MAX", 1, cmdbufsize, cmdbufsize);
+
+	if (attr.test_cmd_max >= 1 && attr.test_cmd_max < cmdbufsize)
+		cmdbufsize = attr.test_cmd_max;
 
 	while (*src)
 	{
@@ -1498,28 +1620,6 @@ void visitf_execute(void)
 
 /*
 
-int fcntl_set_fdflag(int fd, int flag);
-
-Shorthand for setting the file descriptor flag, flag, on the file
-descriptor fd, using fcntl. All other file descriptor flags are
-unaffected. On success, returns 0. On error, returns -1 with errno
-set by fcntl with F_GETFD or F_SETFD as the command. The only
-file descriptor flag at time of writing is FD_CLOEXEC.
-
-*/
-
-int fcntl_set_fdflag(int fd, int flag)
-{
-	int flags;
-
-	if ((flags = fcntl(fd, F_GETFD, 0)) == -1)
-		return -1;
-
-	return fcntl(fd, F_SETFD, flags | flag);
-}
-
-/*
-
 int chdir_local(int do_debug);
 
 Perform the first chdir (before fchdir starts happening).
@@ -1534,7 +1634,6 @@ int chdir_local(int do_debug)
 {
 	char *slash_posp, *dir_path;
 	int dot_fd = -1;
-	char *env;
 
 	/* If in a sub-directory (has parent_id and basename) */
 
@@ -1543,13 +1642,10 @@ int chdir_local(int do_debug)
 		if (do_debug)
 			debug(("fchdir(cwd=%s parent_fd=%d) [%.*s]", getcwd(cwdbuf, CMDBUFSIZE), attr.parent_fd, attr.basename - attr.fpath - 1, attr.fpath));
 
-		if ((env = getenv("RAWHIDE_TEST_FCHDIR_FAILURE")))
-		{
-			debug(("RAWHIDE_TEST_FCHDIR_FAILURE=%s", env));
+		if (attr.test_fchdir_failure)
 			errno = EPERM;
-		}
 
-		if ((env && *env == '1') || fchdir(attr.parent_fd) == -1)
+		if (attr.test_fchdir_failure || fchdir(attr.parent_fd) == -1)
 			fatalsys("fchdir %.*s", attr.basename - attr.fpath - 1, attr.fpath);
 	}
 	else /* Starting search directory (no parent_id or basename) */
@@ -1577,15 +1673,10 @@ int chdir_local(int do_debug)
 			if (do_debug)
 				debug(("initial chdir(cwd=%s path=%s) fpath=%s", getcwd(cwdbuf, CMDBUFSIZE), dir_path, attr.fpath));
 
-			if ((env = getenv("RAWHIDE_TEST_CHDIR_FAILURE")))
-			{
-				if (do_debug)
-					debug(("RAWHIDE_TEST_CHDIR_FAILURE=%s", env));
-
+			if (attr.test_chdir_failure)
 				errno = EPERM;
-			}
 
-			if ((env && *env == '1') || chdir(dir_path) == -1)
+			if (attr.test_chdir_failure || chdir(dir_path) == -1)
 				fatalsys("chdir %s", ok(dir_path));
 
 			free(dir_path);
@@ -1710,7 +1801,7 @@ int remove_danger_from_path(void)
 		return setenv("PATH", DEFAULT_PATH, 1);
 	}
 
-	if (!(oldpath = getenv("PATH")))
+	if (!(oldpath = getenv("PATH")) || !*oldpath)
 	{
 		debug_extra(("no path, setting default path"));
 
@@ -1725,7 +1816,6 @@ int remove_danger_from_path(void)
 	src = oldpath;
 	dst = newpath;
 
-#if 1 /* This version is slower (multiple passes) but more convincing */
 	strlcpy(dst, src, strlen(oldpath) + 1);
 
 	for (;;)
@@ -1780,120 +1870,6 @@ int remove_danger_from_path(void)
 
 		break;
 	}
-#else
-	while (*src)
-	{
-		debug_extra(("path %s pos %d c %c", oldpath, src - oldpath, *src));
-		debug_extra(("     %*s%c", src - oldpath, "", '^'));
-
-		if (src == oldpath && src[0] == ':') /* PATH=:/abc -> PATH=/abc */
-		{
-			debug_extra(("[0] PATH=:/abc -> PATH=/abc [+1]"));
-			src += 1;
-
-			continue;
-		}
-
-		if (src == oldpath && src[0] == '.' && src[1] == ':') /* PATH=.:/abc -> PATH=/abc */
-		{
-			debug_extra(("[1] PATH=.:/abc -> PATH=/abc [+2]"));
-			src += 2;
-
-			continue;
-		}
-
-		if (src > oldpath && src[-1] == ':' && src[0] == '.' && src[1] == ':') /* PATH=/abc:.:/def -> PATH=/abc:/def */
-		{
-			debug_extra(("[2] PATH=/abc:.:/def -> PATH=/abc:/def [+2]"));
-			src += 2;
-
-			continue;
-		}
-
-		if (src > oldpath && src[-1] == ':' && src[0] == '.' && src[1] == '\0') /* PATH=/abc:. -> PATH=/abc */
-		{
-			debug_extra(("[3] PATH=/abc:. -> PATH=/abc [+1]"));
-			src += 1; /* Skip trailing . */
-			if (dst > newpath)
-				dst -= 1; /* Remove previous now trailing : */
-
-			continue;
-		}
-
-		if (src[0] == ':' && src[1] == '.' && src[2] == ':') /* PATH=/abc:.:/def -> PATH=/abc:/def */
-		{
-			debug_extra(("[4] PATH=/abc:.:/def -> PATH=/abc:/def [+2]"));
-			src += 2;
-
-			continue;
-		}
-
-		if (src[0] == ':' && src[1] == ':') /* PATH=/abc::/def -> PATH=/abc:/def */
-		{
-			debug_extra(("[5] PATH=/abc::/def -> PATH=/abc:/def [+1]"));
-			src += 1;
-
-			continue;
-		}
-
-		if (src[0] == ':' && src[1] == '.' && src[2] == '\0') /* PATH=/abc:. -> PATH=/abc */
-		{
-			debug_extra(("[6] PATH=/abc:. -> PATH=/abc [+2]"));
-			src += 2;
-
-			continue;
-		}
-
-		if (src[0] == ':' && src[1] == '\0') /* PATH=/abc: -> PATH=/abc */
-		{
-			debug_extra(("[7] PATH=/abc: -> PATH=/abc [+1]"));
-			src += 1;
-
-			continue;
-		}
-
-		if (src == oldpath && src[0] != '/') /* PATH=rel:/abc -> PATH=/abc */
-		{
-			debug_extra(("[8] PATH=rel:/abc -> PATH=/abc [+len+1]"));
-
-			while (*++src && *src != ':')
-				;
-
-			if (*src == ':')
-				++src;
-
-			continue;
-		}
-
-		if (src >= oldpath && src[-1] == ':' && src[0] != '/' && src[0] != ':') /* PATH=rel:/abc -> PATH=/abc */
-		{
-			debug_extra(("[9] PATH=/abc:rel:/def -> PATH=/abc:/def [+len+1]"));
-
-			while (*++src && *src != ':')
-				;
-
-			if (*src == ':')
-				++src;
-
-			continue;
-		}
-
-		if (src[0] == ':' && src[1] != '/') /* PATH=/abc:rel:/def -> PATH=/abc:/def */
-		{
-			debug_extra(("[10] PATH=/abc:rel:/def -> PATH=/abc:/def [+len+1]"));
-
-			while (*++src && *src != ':')
-				;
-
-			continue;
-		}
-
-		debug_extra(("copy %c", *src));
-		*dst++ = *src++;
-	}
-
-	*dst = '\0';
-#endif
 
 	/* If we no longer have a path, use the default */
 
@@ -1977,6 +1953,302 @@ void visitf_unlink(void)
 
 /*
 
+static int add_field(char *buf, ssize_t sz, const char *name, const char *value);
+
+Add a JSON string field to buf (of sz bytes).
+Encode the value with C-like escape sequences (but without "\a").
+Return the length in bytes (excluding nul).
+
+*/
+
+static int add_field(char *buf, ssize_t sz, const char *name, const char *value)
+{
+	int pos = 0;
+
+	pos += ssnprintf(buf + pos, sz - pos, "\"%s\":\"", name);
+	pos += cescape(buf + pos, sz - pos, value, -1, CESCAPE_JSON);
+	pos += ssnprintf(buf + pos, sz - pos, "\", ");
+
+	return pos;
+}
+
+/*
+
+static char *attributes(void);
+
+Return the ext2-style file attributes as a space-separated
+list of attribute names.
+
+*/
+
+static char *attributes(void)
+{
+	#define ATTR_BUFSIZE 256
+	static char buf[ATTR_BUFSIZE];
+	unsigned long flags = get_attr();
+	int pos = 0;
+
+	buf[pos] = '\0';
+
+	if (attr.test_attr_format)
+		flags = (unsigned long)-1;
+
+	if (!flags)
+		return buf;
+
+	#ifdef EXT2_SECRM_FL
+	if (flags & EXT2_SECRM_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "secrm");
+	#endif
+
+	#ifdef EXT2_UNRM_FL
+	if (flags & EXT2_UNRM_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "unrm");
+	#endif
+
+	#ifdef EXT2_COMPR_FL
+	if (flags & EXT2_COMPR_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "compr");
+	#endif
+
+	#ifdef EXT2_SYNC_FL
+	if (flags & EXT2_SYNC_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "sync");
+	#endif
+
+	#ifdef EXT2_IMMUTABLE_FL
+	if (flags & EXT2_IMMUTABLE_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "immutable");
+	#endif
+
+	#ifdef EXT2_APPEND_FL
+	if (flags & EXT2_APPEND_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "append");
+	#endif
+
+	#ifdef EXT2_NODUMP_FL
+	if (flags & EXT2_NODUMP_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "nodump");
+	#endif
+
+	#ifdef EXT2_NOATIME_FL
+	if (flags & EXT2_NOATIME_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "noatime");
+	#endif
+
+	#ifdef EXT2_DIRTY_FL
+	if (flags & EXT2_DIRTY_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "dirty");
+	#endif
+
+	#ifdef EXT2_COMPRBLK_FL
+	if (flags & EXT2_COMPRBLK_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "comprblk");
+	#endif
+
+	#ifdef EXT2_NOCOMPR_FL
+	if (flags & EXT2_NOCOMPR_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "nocompr");
+	#endif
+
+	#ifdef EXT4_ENCRYPT_FL
+	if (flags & EXT4_ENCRYPT_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "encrypt");
+	#endif
+
+	#ifdef EXT2_INDEX_FL
+	if (flags & EXT2_INDEX_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "index");
+	#endif
+
+	#ifdef EXT2_IMAGIC_FL
+	if (flags & EXT2_IMAGIC_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "imagic");
+	#endif
+
+	#ifdef EXT3_JOURNAL_DATA_FL
+	if (flags & EXT3_JOURNAL_DATA_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "journal_data");
+	#endif
+
+	#ifdef EXT2_NOTAIL_FL
+	if (flags & EXT2_NOTAIL_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "notail");
+	#endif
+
+	#ifdef EXT2_DIRSYNC_FL
+	if (flags & EXT2_DIRSYNC_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "dirsync");
+	#endif
+
+	#ifdef EXT2_TOPDIR_FL
+	if (flags & EXT2_TOPDIR_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "topdir");
+	#endif
+
+	#ifdef EXT4_HUGE_FILE_FL
+	if (flags & EXT4_HUGE_FILE_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "huge_file");
+	#endif
+
+	#ifdef EXT4_EXTENTS_FL
+	if (flags & EXT4_EXTENTS_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "extents");
+	#endif
+
+	#ifdef EXT4_VERITY_FL
+	if (flags & EXT4_VERITY_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "verity");
+	#endif
+
+	#ifdef EXT4_EA_INODE_FL
+	if (flags & EXT4_EA_INODE_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "ea_inode");
+	#endif
+
+	#ifdef FS_NOCOW_FL
+	if (flags & FS_NOCOW_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "nocow");
+	#endif
+
+	#ifdef EXT4_SNAPFILE_FL
+	if (flags & EXT4_SNAPFILE_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "snapfile");
+	#endif
+
+	#ifdef FS_DAX_FL
+	if (flags & FS_DAX_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "dax");
+	#endif
+
+	#ifdef EXT4_SNAPFILE_DELETED_FL
+	if (flags & EXT4_SNAPFILE_DELETED_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "snapfile_deleted");
+	#endif
+
+	#ifdef EXT4_SNAPFILE_SHRUNK_FL
+	if (flags & EXT4_SNAPFILE_SHRUNK_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "snapfile_shrunk");
+	#endif
+
+	#ifdef EXT4_INLINE_DATA_FL
+	if (flags & EXT4_INLINE_DATA_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "inline_data");
+	#endif
+
+	#ifdef EXT4_PROJINHERIT_FL
+	if (flags & EXT4_PROJINHERIT_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "projinherit");
+	#endif
+
+	#ifdef EXT4_CASEFOLD_FL
+	if (flags & EXT4_CASEFOLD_FL)
+		pos += ssnprintf(buf + pos, ATTR_BUFSIZE - pos, "%s%s", (pos) ? " " : "", "casefold");
+	#endif
+
+	return buf;
+}
+
+/*
+
+static char *json(void);
+
+Return all file information in JSON format. The information
+included is path, name, target path, starting path, depth,
+and all the inode metadata (stat(2) structure fields).
+
+*/
+
+static char *json(void)
+{
+	#define JSON_BUFSIZE 262144
+	static char buf[JSON_BUFSIZE];
+	struct passwd *pwd;
+	struct group *grp;
+	char *base, *selinux, *acl, *ea;
+	int pos = 0;
+
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "{");
+
+	pos += add_field(buf + pos, JSON_BUFSIZE - pos, "path", attr.fpath);
+
+	base = ((base = strrchr(attr.fpath, '/')) && base[1]) ? base + 1 : attr.fpath;
+	pos += add_field(buf + pos, JSON_BUFSIZE - pos, "name", base);
+
+	if (islink(attr.statbuf))
+		pos += add_field(buf + pos, JSON_BUFSIZE - pos, "target", read_symlink());
+
+	pos += add_field(buf + pos, JSON_BUFSIZE - pos, "start", attr.search_path);
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"depth\":%lld, ", (llong)attr.depth);
+
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"dev\":%lld, ", (llong)attr.statbuf->st_dev);
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"major\":%lld, ", (llong)major(attr.statbuf->st_dev));
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"minor\":%lld, ", (llong)minor(attr.statbuf->st_dev));
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"ino\":%lld, ", (llong)attr.statbuf->st_ino);
+
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"mode\":%lld, ", (llong)attr.statbuf->st_mode);
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"modestr\":\"%s\", ", modestr(attr.statbuf));
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"type\":\"%s\", ", ytypecode(attr.statbuf));
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"perm\":%d, ", (int)attr.statbuf->st_mode & ~S_IFMT);
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"nlink\":%lld, ", (llong)attr.statbuf->st_nlink);
+
+	if ((pwd = getpwuid(attr.statbuf->st_uid)))
+		pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"user\":\"%s\", ", pwd->pw_name);
+
+	if ((grp = getgrgid(attr.statbuf->st_gid)))
+		pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"group\":\"%s\", ", grp->gr_name);
+
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"uid\":%lld, ", (llong)attr.statbuf->st_uid);
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"gid\":%lld, ", (llong)attr.statbuf->st_uid);
+
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"rdev\":%lld, ", (llong)attr.statbuf->st_rdev);
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"rmajor\":%lld, ", (llong)major(attr.statbuf->st_rdev));
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"rminor\":%lld, ", (llong)minor(attr.statbuf->st_rdev));
+
+	set_dirsize();
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"size\":%lld, ", (llong)attr.statbuf->st_size);
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"blksize\":%lld, ", (llong)attr.statbuf->st_blksize);
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"blocks\":%lld, ", (llong)attr.statbuf->st_blocks);
+
+	pos += strftime(buf + pos, JSON_BUFSIZE - pos, "\"atime\":\"%Y-%m-%d %H:%M:%S %z\", ", localtime(&attr.statbuf->st_atime));
+	pos += strftime(buf + pos, JSON_BUFSIZE - pos, "\"mtime\":\"%Y-%m-%d %H:%M:%S %z\", ", localtime(&attr.statbuf->st_mtime));
+	pos += strftime(buf + pos, JSON_BUFSIZE - pos, "\"ctime\":\"%Y-%m-%d %H:%M:%S %z\", ", localtime(&attr.statbuf->st_ctime));
+
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"atime_unix\":%lld, ", (llong)attr.statbuf->st_atime);
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"mtime_unix\":%lld, ", (llong)attr.statbuf->st_mtime);
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"ctime_unix\":%lld, ", (llong)attr.statbuf->st_ctime);
+
+	#ifdef HAVE_ATTR
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"attributes\":\"%s\", ", attributes());
+
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"project\":%lu, ", get_proj());
+
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"generation\":%lu, ", get_gen());
+	#endif
+
+	if ((acl = get_acl(1)))
+	{
+		pos += add_field(buf + pos, JSON_BUFSIZE - pos, "access_control_list", acl);
+
+		if (attr.facl_verbose)
+			pos += add_field(buf + pos, JSON_BUFSIZE - pos, "access_control_list_verbose", acl);
+	}
+
+	if ((ea = get_ea(1)) && attr.fea_ok)
+		pos += add_field(buf + pos, JSON_BUFSIZE - pos, "extended_attributes", ea);
+
+	if ((selinux = (get_ea(1) && attr.fea_ok && attr.fea_selinux) ? attr.fea_selinux : "") && *selinux)
+		pos += add_field(buf + pos, JSON_BUFSIZE - pos, "selinux_context", selinux);
+
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "\"acl_ea_indicator\":\"%s\"", aclea());
+
+	pos += ssnprintf(buf + pos, JSON_BUFSIZE - pos, "}");
+
+	return buf;
+}
+
+/*
+
 void visitf_format(void);
 
 The -L action for outputting match information in a user-supplied format.
@@ -2625,6 +2897,39 @@ void visitf_format(void)
 						break;
 					}
 
+					case 'e': /* Ext2-style file attributes */
+					{
+						char *a = attributes();
+
+						ofmt_add('s');
+						debug_extra(("fmt %%e \"%s\", \"%s\"", ofmt, a));
+						printf(ofmt, a);
+
+						break;
+					}
+
+					case 'J': /* Ext2-style project */
+					{
+						unsigned long proj = get_proj();
+
+						ofmt_add_lld();
+						debug_extra(("fmt %%J \"%s\", %lld", ofmt, (llong)proj));
+						printf(ofmt, (llong)proj);
+
+						break;
+					}
+
+					case 'I': /* Ext2-style generation */
+					{
+						unsigned long gen = get_gen();
+
+						ofmt_add_lld();
+						debug_extra(("fmt %%I \"%s\", %lld", ofmt, (llong)gen));
+						printf(ofmt, (llong)gen);
+
+						break;
+					}
+
 					case '\0': /* Invalid % at the end */
 					{
 						fatal("invalid -L argument: %s (%% at the end)", ok(attr.format));
