@@ -2,6 +2,7 @@
 * rawhide - find files using pretty C expressions
 * https://raf.org/rawhide
 * https://github.com/raforg/rawhide
+* https://codeberg.org/raforg/rawhide
 *
 * Copyright (C) 1990 Ken Stauffer, 2022 raf <raf@raf.org>
 *
@@ -18,7 +19,7 @@
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <https://www.gnu.org/licenses/>.
 *
-* 20220330 raf <raf@raf.org>
+* 20221011 raf <raf@raf.org>
 */
 
 #ifndef RAWHIDE_RHDIR_H
@@ -27,10 +28,7 @@
 llong env_int(char *envname, llong min_value, llong max_value, llong default_value);
 int rawhide_search(char *fpath);
 int following_symlinks(void);
-const char *ytypecode(struct stat *statbuf);
 const char *modestr(struct stat *statbuf);
-const char *aclea(void);
-int fcntl_set_fdflag(int fd, int flag);
 void visitf_default(void);
 void visitf_long(void);
 void visitf_execute(void);
