@@ -2,6 +2,7 @@
 * rawhide - find files using pretty C expressions
 * https://raf.org/rawhide
 * https://github.com/raforg/rawhide
+* https://codeberg.org/raforg/rawhide
 *
 * Copyright (C) 1990 Ken Stauffer, 2022 raf <raf@raf.org>
 *
@@ -18,7 +19,7 @@
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <https://www.gnu.org/licenses/>.
 *
-* 20220330 raf <raf@raf.org>
+* 20221011 raf <raf@raf.org>
 */
 
 #define _GNU_SOURCE /* For FNM_EXTMATCH and FNM_CASEFOLD in <fnmatch.h> */
@@ -84,7 +85,7 @@ static const char *show_token(void);
 
 static void debugf(const char *format, ...);
 
-Print a debug message to stderr.
+Output a parser debug message to stderr, if requested.
 
 */
 
@@ -106,7 +107,7 @@ static void debugf(const char *format, ...)
 
 static void debug_extraf(const char *format, ...);
 
-Print an extra debug message to stderr.
+Ouptut an extra parser debug message to stderr, if requested.
 
 */
 
@@ -188,7 +189,7 @@ static void ungetch(int c)
 
 static void parser_error(const char *format, ...);
 
-Print a syntax message and exit.
+Output a syntax error message and exit.
 The format parameter is a printf-like format.
 Subsequent arguments must satisfy its conversions.
 
@@ -1068,7 +1069,6 @@ static llong parse_datetime(void)
 	int year, month, day, hour = 0, minute = 0, second = 0;
 	llong seconds;
 	struct tm tm[1];
-	char *env;
 
 	debug_extra(("datetime()"));
 
@@ -1147,7 +1147,7 @@ static llong parse_datetime(void)
 	tm->tm_yday = 0;
 	tm->tm_isdst = -1;
 
-	if ((seconds = (llong)mktime(tm)) == -1 || (year == 1900 && (env = getenv("RAWHIDE_TEST_INVALID_DATE")) && *env))
+	if ((seconds = (llong)mktime(tm)) == -1 || (year == 1900 && attr.test_invalid_date))
 	{
 		if (hour == 0 && minute == 0 && second == 0)
 			parser_error("invalid date: [%04d/%d/%d]", year, month, day);
@@ -1286,7 +1286,7 @@ static int _get_token(void)
 		c = getch();
 	}
 
-	/* Octal/Hexadecimal */
+	/* Octal / Hexadecimal */
 
 	if (c == '0' && !decimal_mode)
 	{
@@ -1432,7 +1432,7 @@ static int _get_token(void)
 			return PATMOD;
 		}
 
-		/* Get the reference file field name */
+		/* Get the pattern modifier or reference file field name */
 
 		reffield = refstrfree = strfree;
 
@@ -1452,11 +1452,34 @@ static int _get_token(void)
 		ungetch(c);
 		Strbuf[refstrfree] = '\0';
 
-		if (!(tokensym = locate_symbol(Strbuf + reffield)) || (tokensym->type != REFFILE && tokensym->type != PATMOD))
+		/* Look for the symbol (it can be a unique prefix of a pattern modifier) */
+
+		if (!(tokensym = locate_symbol(Strbuf + reffield)))
+			tokensym = locate_patmod_prefix(Strbuf + reffield);
+
+		if (!tokensym || (tokensym->type != REFFILE && tokensym->type != PATMOD))
 			parser_error("invalid string suffix: %s (expected pattern modifier or reference file field)", ok(Strbuf + reffield));
 
+		/* When / is present, replace i with ipath, re with repath, and rei with reipath */
+
 		if (tokensym->type == PATMOD)
+		{
+			if (strchr(Strbuf + tokenval, '/'))
+			{
+				#ifdef FNM_CASEFOLD
+				if (!strcmp(tokensym->name, ".i"))
+					tokensym = locate_symbol(".ipath");
+				#endif
+				#ifdef HAVE_PCRE2
+				if (!strcmp(tokensym->name, ".re"))
+					tokensym = locate_symbol(".repath");
+				else if (!strcmp(tokensym->name, ".rei"))
+					tokensym = locate_symbol(".reipath");
+				#endif
+			}
+
 			return PATMOD;
+		}
 
 		/* Strip trailing / */
 
@@ -1666,7 +1689,7 @@ static int _get_token(void)
 static int get_token_decimal(void);
 
 Like get_token() with decimal_mode set first and cleared afterwards.
-This is for parsing datetimes to allow leading zeroes.
+This is for parsing datetimes to allow leading zeroes for decimals.
 
 */
 
